      subroutine ev_test(N, NVEC, A, lda, w, z, ldz)
      use MPI
!     $    use OMP_LIB
      use eigen_libs_mod
      use eigen_blacs_mod
      implicit none

      integer,intent(in) :: N, NVEC
      real(8),intent(in) :: w(*)
      real(8),intent(inout) :: z(*)
      real(8),intent(in) :: A(*)
      integer,intent(in) :: lda,ldz

      real(8),parameter :: one = 1.0D0
      real(8),parameter :: zero = 0.0D0

      integer :: NPROW,NPCOL,larray
      integer :: ictxt
      integer :: iam1,nprocs
      integer :: myrow1,mycol1
      integer :: numr_row,numr_col
      integer :: ierr,inod
      integer :: i,j
      real(8) :: s_time,e_time
      integer :: nm, nx, NB

      integer :: numroc,iceil,indxl2g
      external :: numroc,iceil,indxl2g
      real(8) :: pdlange
      external :: pdlange

      real(8) :: err1, err2, rr, m_epsilon
      integer :: gcol
      real(8) :: DUMWORK
      real(8), allocatable :: Y(:)
      real(8) :: buf(6)

      integer, parameter :: DESC_DIM = 9
      integer, dimension(DESC_DIM) :: a0desc, z0desc
      integer, dimension(DESC_DIM) :: a1desc, z1desc

      integer :: i_1,i_2,i_3
      integer :: j_1,j_2,j_3
      integer :: iam, myrow, mycol      


      m_epsilon = get_constant_eps()

      call eigen_get_id   ( iam, myrow, mycol )            
      call eigen_get_procs( nprocs, NPROW, NPCOL )            
      inod=iam-1; myrow1=myrow-1; mycol1=mycol-1

      call eigen_get_matdims( n, nm, nx )

      larray = nm * nx
*     
      NB = eigen_NB
      nm = (((n-1)/NB)/NPROW+1)*NB
      nx = (((NVEC-1)/NB)/NPCOL+1)*NB

      larray = max( larray, nm * nx )

      ictxt = eigen_get_blacs_context()
      call descinit( a0desc, n, n, 1, 1, 0, 0, ictxt, lda, ierr )
      call descinit( z0desc, n, n, 1, 1, 0, 0, ictxt, ldz, ierr )
      call descinit( a1desc, n, n, NB, NB, 0, 0, ictxt, nm, ierr )
      call descinit( z1desc, n, n, NB, NB, 0, 0, ictxt, nm, ierr )
*
* local buffers are elimibnated by sharing other arrays, which
* are allocated largely according to the get_matdims mode='O'.
*
      allocate(Y(larray))
!
! a ==> Y
! z ==> A
!
      call pdgemr2d( n, n, a, 1, 1, a0desc,
     &     Y, 1, 1, a1desc, ictxt)
      call pdgemr2d( n, n, z, 1, 1, z0desc,
     &     A, 1, 1, z1desc, ictxt)

! 1. |AZ-ZW|
!    Z = YA = az
      call pdgemm('N', 'N', n, NVEC, n,
     &     one,
     &     Y, 1, 1, a1desc,
     &     A, 1, 1, z1desc,
     &     zero,
     &     Z, 1, 1, z1desc)

!    rr = \| Y \|_{F} = \| a \|_{F}
      rr = Fnorm2_local( n, n, Y(1:), nm, NB )
!      rr = pdlange( 'F', n, n, Y, 1, 1, z1desc, DUMWORK )
      deallocate( Y )

!    Z = Z - WA = az - zw
!$OMP PARALLEL DO PRIVATE(i,j_1,j_2)
      do i_1=1, nx
         j_1 = MOD(i_1-1,NB)+1
         j_2 = (i_1-1)/NB*NPCOL+(MYCOL-1)
         i   = j_1 + j_2*NB
         j_1 = 1 + (i_1-1)*nm
         call DAXPY(nm, -w(i), A(j_1), 1, Z(j_1), 1)
      end do
!$OMP END PARALLEL DO
!    err1 = \| Z \|_{F} = \| az-zw \|_{F}
      err1 = Fnorm2_local( n, NVEC, Z, nm, NB )
!      err1 = pdlange( 'F', n, NVEC, Z, 1, 1, z1desc, DUMWORK )

! 2. |ZZ-I|
!    Z = AA = zz
      call pdgemm('T', 'N', NVEC, NVEC, n,
     &     one,
     &     A, 1, 1, z1desc,
     &     A, 1, 1, z1desc,
     &     zero,
     &     Z, 1, 1, z1desc)

!    Z = Z - eye = zz - eye
!$OMP PARALLEL DO PRIVATE(i,j,j_1,j_2)
      do i_1=1,nx
         j_1 = MOD(i_1-1,NB)+1
         j_2 = (i_1-1)/NB*NPCOL+(MYCOL-1)
         j   = MOD(j_2,NPROW)+1
         if (j == MYROW) then
            j_1 = (j_1 + (j_2/NPROW)*NB) + (i_1-1)*nm
            Z(j_1) = Z(j_1) - one
         end if
      end do
!$OMP END PARALLEL DO
!    err2 = \| Z \|_{F} = \| zz-eye \|_{F}
      err2 = Fnorm2_local( NVEC, NVEC, Z, nm, NB )
!      err2 = pdlange( 'F', NVEC, NVEC, Z, 1, 1, z1desc, DUMWORK )

      buf(1) = rr
      buf(2) = err1
      buf(3) = err2
      call MPI_Allreduce(buf(1), buf(4), 3, MPI_DOUBLE_PRECISION,
     &       MPI_SUM, MPI_COMM_WORLD, ierr)
      rr   = sqrt(buf(4))
      err1 = sqrt(buf(5))
      err2 = sqrt(buf(6))

      if ( inod == 0 ) then

         rr = err1/(dble(n)*rr*m_epsilon)
         print*,'|AZ-ZW|_{F}/Ne|A|_{F}=', rr
         if ( rr < sqrt(8d0) ) then
          print*,"*** Residual Error Test ***   : PASSED"
         else
          print*,"***=============================******"
          print*,"*** Residual Error Test ***   : FAILED"
          print*,"***=============================******"
         end if

         rr = err2/(dble(n)*m_epsilon)
         print*,'|ZZ-I|_{F}/Ne=        ', rr
         if ( rr < sqrt(8d0) ) then
          print*,"*** Orthogonality  Test ***   : PASSED"
         else
          print*,"***=============================******"
          print*,"*** Orthogonality  Test ***   : FAILED"
          print*,"***=============================******"
         end if

      end if

      return
      contains

      real(8) function Fnorm2_local( M, N, A, LDA, NB )
      use MPI
!$    use OMP_LIB
      integer, intent(IN) :: M, N, LDA, NB
      real(8), intent(IN) :: A(1:LDA,1:*)
!
      integer :: lx, ly, i, j, nth
      real(8) :: r
      real(8), pointer :: rr(:)
!
!
!     find i such that;
!       (MOD(i-1,NB)+1)+((i-1)/NB*NPCOL+MYCOL)*NB <= N
!        MOD(i-1,NB)+((i-1)/NB*NPCOL)*NB <= N-MYCOL*NB-1
!     if i-1:=i0+i1*NB, 0<=i0<NB, 0<=i1
!        i0 <= MOD(N-MYCOL*NB-1,NB)
!        i1*NPCOL <= (N-MYCOL*NB-1)/NB
!     =>
!        i0 <= MOD(N-MYCOL*NB-1,NB)
!        i1 <= (N-MYCOL*NB-1)/NB/NPCOL
!
      i = max(0, N - MYCOL1*NB -1)
      j = max(0, M - MYROW1*NB -1)
      ly = MOD(i, NB) + ((i/NB)/NPCOL)*NB
      lx = MOD(j, NB) + ((j/NB)/NPROW)*NB

!$OMP PARALLEL PRIVATE(i,j,r)
!$OMP MASTER
!$    allocate( rr(1:omp_get_num_threads()) )
!$OMP END MASTER
!$OMP BARRIER

      r = zero
!$OMP DO COLLAPSE(2)
      do i=1,ly
        do j=1,lx
          r = r + A(j,i)**2
        end do
      end do
!$OMP END DO
!$    rr(omp_get_thread_num()+1) = r
!$OMP BARRIER

!$OMP MASTER
!$    do i=2, omp_get_num_threads()
!$      r = r + rr(i)
!$    end do
!$    deallocate( rr )
      Fnorm2_local = r
!$OMP END MASTER
!$OMP END PARALLEL

      return
      end function Fnorm2_local

      end subroutine ev_test

